import dataclasses
import inspect
from collections.abc import Callable, Mapping, Sequence
from types import EllipsisType, NoneType
from typing import Annotated, Concatenate, Generic, TypeAlias, get_origin, get_type_hints, overload

from docstring_parser import parse
from docstring_parser.common import DocstringParam
from pydantic import BaseModel, Field, ValidationError, create_model
from typing_extensions import NotRequired, ParamSpec, TypedDict, TypeVar


class TaskEnvironment(BaseModel):
    """Base class for the environment in which the agent is running."""

    ...


class EmptyEnv(TaskEnvironment):
    """An empty environment for when a state is not needed."""

    ...


@dataclasses.dataclass(frozen=True)
class Depends:
    """A dependency for a function. It is used to extract information from the environment to pass to the function."""

    env_dependency: str | Callable[[BaseModel], BaseModel]
    """The name of the attribute to extract from the environment or a function that returns the attribute from the environment."""

    def extract_dep_from_env(self, env: BaseModel) -> BaseModel:
        """Extracts the dependency from the environment."""
        if callable(self.env_dependency):
            return self.env_dependency(env)
        return getattr(env, self.env_dependency)

    def __repr__(self) -> str:
        return f"Depends('{self.env_dependency}')"


FunctionCallArgTypes = str | int | float | bool | NoneType | dict | list | EllipsisType
"""Valid types for function call arguments."""
FunctionCallArgs = Mapping[str, FunctionCallArgTypes]
"""Type for a dictionary of function call arguments."""


class FunctionCall(TypedDict):
    """A dictionary containing information about a function call requested by an agent."""

    function: str
    """The name of the function to call."""
    args: FunctionCallArgs
    """The arguments to pass to the function."""
    id: NotRequired[str]
    """An optional ID for the function call. E.g., used by OpenAI and Anthropic."""
    placeholder_args: NotRequired[FunctionCallArgs]
    """An optional dictionary of placeholder arguments to use in by ground truth agent in injection tasks."""


FunctionReturnType: TypeAlias = BaseModel | Sequence["FunctionReturnType"] | dict | str | int | float | bool | NoneType
P = ParamSpec("P")
S = TypeVar("S")
Env = TypeVar("Env", bound=TaskEnvironment)
T = TypeVar("T", bound=FunctionReturnType)
FunctionCallable = Callable[Concatenate[Env, P], FunctionReturnType]


class Function(BaseModel, Generic[P, T]):
    """A function that can be called by an agent and run by a [`FunctionsRuntime`][agentdojo.functions_runtime.FunctionsRuntime]."""

    name: str
    """The name of the function."""
    description: str
    """A short description of the function."""
    parameters: type[BaseModel]
    """The input schema for the function. It is defined as a Pydantic model."""
    dependencies: dict[str, Depends]
    """A dictionary of dependencies for the function. See [`Depends`][agentdojo.functions_runtime.Depends] for more information."""
    run: Callable[P, T]

    def __call__(self, *args: P.args, **kwargs: P.kwargs) -> T:
        return self.run(*args, **kwargs)


def _get_dependencies(function: Callable[P, T]) -> dict[str, Depends]:
    arg_types = get_type_hints(function, include_extras=True)
    dependencies: dict[str, Depends] = {}
    for arg_name, arg_type in arg_types.items():
        if get_origin(arg_type) is Annotated:
            if isinstance(dependency := arg_type.__metadata__[0], Depends):
                dependencies[arg_name] = dependency
    return dependencies


def make_function(function: Callable[P, T]) -> Function:
    """Creates a [`Function`][agentdojo.functions_runtime.Function] instance from a function.

    It does so by extracting the docstrings and type hints from the function. It then creates
    a Pydantic model for the input schema of the function.

    Note:
        ReStructuredText is the recommended docstring format.

    Args:
        function: The function to create a [`Function`][agentdojo.functions_runtime.Function] instance from.
            The function should have docstrings with a short description of the function and descriptions
            for each argument. [`Depends`][agentdojo.functions_runtime.Depends] arguments should not be
            documented in the docstring.

    Returns:
        A [`Function`][agentdojo.functions_runtime.Function] instance.

    Raises:
        ValueError: If the function has no docstring or short description, or if the arguments docstrings are
            invalid.
    """
    dependencies = _get_dependencies(function)

    if function.__doc__ is None or function.__doc__ == "":
        raise ValueError(f"Function {function.__name__} has no docstring")
    function_docs = parse(function.__doc__)
    if function_docs.short_description is None:
        raise ValueError(f"Function {function.__name__} has no short description")
    return Function(
        name=function.__name__,
        description=function_docs.short_description.replace("None", "null"),
        parameters=_parse_args(function.__name__, function_docs.params, inspect.signature(function)),
        run=function,
        dependencies=dependencies,
    )


BM = TypeVar("BM", bound=BaseModel)


def _parse_args(
    function_name: str,
    params_docs: list[DocstringParam],
    signature: inspect.Signature,
    field_fn: Callable = Field,
    create_model_fn: Callable[..., type[BM]] = create_model,
) -> type[BM]:
    params = {}
    for arg in params_docs:
        arg_name = arg.arg_name
        if arg.description is None:
            raise ValueError(f"Argument {arg_name} has no description")
        arg_type = signature.parameters[arg_name].annotation
        default_value = signature.parameters[arg_name].default
        params[arg_name] = (
            arg_type,
            field_fn(
                description=arg.description,
                default=... if default_value == inspect.Parameter.empty else default_value,
            ),
        )

    return create_model_fn(f"Input schema for `{function_name}`", **params)


class ToolNotFoundError(Exception):
    """Exception raised when the requested tool is not found."""

    ...


class FunctionsRuntime:
    """A runtime for functions.

    Args:
        functions: A list of functions that the runtime has access to.
    """

    def __init__(self, functions: Sequence[Function] = []) -> None:
        self.functions: dict[str, Function] = {f.name: f for f in functions}

    def __repr__(self) -> str:
        return f"FunctionsRuntime(functions={self.functions})"

    @overload
    def register_function(self, function: Callable[P, T]) -> Callable[P, T]: ...
    @overload
    def register_function(self, function: Function) -> Function: ...

    def register_function(self, function: Callable[P, T] | Function) -> Callable[P, T] | Function:
        """Registers a function with the runtime.

        If `function` is a function, it will be converted to a `Function` instance and registered,
        otherwise it will be registered as is. The conversion is made with [`make_function`][agentdojo.functions_runtime.make_function].

        Args:
            function: The function to register. It can be a function or a `Function` instance. See [`make_function`][agentdojo.functions_runtime.make_function]
                for more information.

        Returns:
            The registered function.
        """
        if isinstance(function, Function):
            self.functions[function.name] = function
            return function
        try:
            f = make_function(function)
        except Exception as e:
            raise Exception(f"Error when registering function {function.__name__} in {function.__module__}.") from e
        self.functions[function.__name__] = f
        return function

    def run_function(
        self, env: TaskEnvironment | None, function: str, kwargs: FunctionCallArgs, raise_on_error: bool = False
    ) -> tuple[FunctionReturnType, str | None]:
        """Runs a function with the given arguments.

        If `raise_on_error` is `True`, the function will raise an exception if an error occurs. Otherwise, it will return an error message as the second
        argument. The error message has the format `'ErrorType: ErrorMessage'`.

        Args:
            env: The environment to extract dependencies from. It can be `None` if the function has no dependencies.
            function: The name of the function to run.
            kwargs: The arguments to pass to the function.
            raise_on_error: If `True`, the function will raise an exception if an error occurs. Otherwise, it will return an error message.

        Returns:
            A tuple containing the result of the function and an error message if an error occurred. If no error occurred, the second element will be `None`.

        Raises:
            ValidationError: If the arguments are invalid and do not match the function's input schema.
            ToolNotFoundError: If the requested function is not available.
        """
        if function not in self.functions:
            if raise_on_error:
                raise ToolNotFoundError(f"The requested function `{function}` is not available.")
            return "", f"ToolNotFoundError: The requested function `{function}` is not available."
        f = self.functions[function]
        try:
            args = f.parameters.model_validate(kwargs).model_dump()
        except ValidationError as e:
            if raise_on_error:
                raise e
            return "", f"ValidationError: {e}"

        if env is None and len(f.dependencies) > 0:
            raise ValueError(f"Function {function} has dependencies but no environment was provided.")
        else:
            env_args = {
                arg_name: dependency.extract_dep_from_env(env)  # type: ignore -- If there are no dependencies, this line will not be executed
                for arg_name, dependency in f.dependencies.items()
            }
        kwargs_with_deps = args | env_args

        try:
            return f(**kwargs_with_deps), None
        except Exception as e:
            if raise_on_error:
                raise e
            return "", f"{type(e).__name__}: {e}"

    def update_functions(self, new_functions: dict[str, Function]) -> None:
        self.functions = new_functions



from agentdojo.default_suites.v1.travel.injection_tasks import (
    InjectionTask0,
    InjectionTask1,
    InjectionTask2,
    InjectionTask3,
    InjectionTask4,
    InjectionTask5,
    InjectionTask6,
)
from agentdojo.default_suites.v1.travel.task_suite import TravelEnvironment
from agentdojo.default_suites.v1.travel.task_suite import task_suite as travel_task_suite_v1
from agentdojo.task_suite.task_suite import TaskSuite

task_suite = TaskSuite[TravelEnvironment]("travel", TravelEnvironment, travel_task_suite_v1.tools)

injection_tasks = [
    InjectionTask0,
    InjectionTask1,
    InjectionTask2,
    InjectionTask3,
    InjectionTask4,
    InjectionTask5,
    InjectionTask6,
]

for inj_task in injection_tasks:
    task_suite.register_injection_task(inj_task)



import datetime
from typing import Annotated

from pydantic import BaseModel, EmailStr, computed_field, model_validator
from typing_extensions import Self

from agentdojo.default_suites.v1.tools.types import CalendarEvent, CloudDriveFileID, Email, EmailID, EmailStatus
from agentdojo.functions_runtime import Depends


class EmailContact(BaseModel):
    email: EmailStr
    name: str


class Inbox(BaseModel):
    account_email: EmailStr
    initial_emails: list[Email]
    emails: dict[EmailID, Email] = {}
    contact_list: list[EmailContact] = []
    trash: dict[EmailID, Email] = {}

    @model_validator(mode="after")
    def _create_emails(self) -> Self:
        all_ids = [email.id_ for email in self.initial_emails]
        if len(all_ids) != len(set(all_ids)):
            duplicate_ids = [id_ for id_ in all_ids if all_ids.count(id_) > 1]
            raise ValueError(f"Email IDs must be unique. Duplicates: {duplicate_ids}")
        self.emails = {email.id_: email for email in self.initial_emails}
        return self

    @model_validator(mode="after")
    def _create_contact_list(self) -> Self:
        existing_contacts = set()
        for email in self.initial_emails:
            email_contacts = [*email.recipients, email.sender, *email.cc, *email.bcc]
            for recipient in email_contacts:
                if recipient in existing_contacts:
                    continue
                if recipient not in [contact.email for contact in self.contact_list]:
                    name = recipient.split("@")[0]
                    name = name.replace(".", " ").title()
                    new_contact = EmailContact(email=recipient, name=name)
                    self.contact_list.append(new_contact)
                    existing_contacts.add(recipient)
        return self

    def from_address(self, address: str) -> list[Email]:
        return [email for email in self.emails.values() if email.sender == address]

    def _get_next_id(self) -> EmailID:
        largest_id = max((int(key) for key in self.emails.keys()), default=0)
        return EmailID(largest_id + 1)

    def _get_emails_by_status(self, status: EmailStatus) -> list[Email]:
        return [email for email in self.emails.values() if email.status == status]

    @computed_field
    @property
    def received(self) -> list[Email]:
        return self._get_emails_by_status(EmailStatus.received)

    @computed_field
    @property
    def sent(self) -> list[Email]:
        return self._get_emails_by_status(EmailStatus.sent)

    @computed_field
    @property
    def drafts(self) -> list[Email]:
        return self._get_emails_by_status(EmailStatus.draft)

    def send_email(
        self,
        recipients: list[EmailStr],
        subject: str,
        body: str,
        attachments: list[CloudDriveFileID | CalendarEvent] | None = None,
        cc: list[EmailStr] | None = None,
        bcc: list[EmailStr] | None = None,
    ) -> Email:
        if cc is None:
            cc = []
        if bcc is None:
            bcc = []
        if attachments is None:
            attachments = []
        new_email = Email(
            id_=self._get_next_id(),
            sender=self.account_email,
            body=body,
            subject=subject,
            recipients=recipients,
            status=EmailStatus.sent,
            timestamp=datetime.datetime.now(),
            cc=cc,
            bcc=bcc,
            attachments=attachments,
            read=True,
        )
        self.emails[new_email.id_] = new_email
        return new_email

    def search_emails(self, query: str, sender: EmailStr | None = None) -> list[Email]:
        if sender is not None:
            emails = self.from_address(sender)
        else:
            emails = self.emails.values()
        return [
            email for email in emails if query.lower() in email.subject.lower() or query.lower() in email.body.lower()
        ]

    def delete_email(self, email_id: EmailID) -> Email:
        if email_id not in self.emails:
            raise ValueError(f"Email with ID '{email_id}' not found.")
        email = self.emails.pop(email_id)
        self.trash[email_id] = email
        return email

    def get_unread(self) -> list[Email]:
        unread_emails = [email for email in self.emails.values() if not email.read]
        for email in unread_emails:
            email.read = True
        return unread_emails

    def find_contacts_by_name(self, name: str) -> list[EmailContact]:
        matches = []
        for contact in self.contact_list:
            if name.lower() in contact.name.lower():
                matches.append(contact)
        if len(matches) == 0:
            raise ValueError(f"Contact with name '{name}' not found.")
        return matches

    def find_contacts_by_email(self, email: EmailStr) -> list[EmailContact]:
        matches = []
        for contact in self.contact_list:
            if email.lower() in contact.email.lower():
                matches.append(contact)
        if not matches:
            raise ValueError(f"Contact with email '{email}' not found.")
        return matches

    def get_most_recently_received_email(self) -> Email:
        return max(self.received, key=lambda email: email.timestamp)


def get_unread_emails(inbox: Annotated[Inbox, Depends("inbox")]) -> list[Email]:
    """Returns all the unread emails in the inbox. Each email has a sender, a subject, and a body.
    The emails are marked as read after this function is called."""
    return inbox.get_unread()


def send_email(
    inbox: Annotated[Inbox, Depends("inbox")],
    recipients: list[str],
    subject: str,
    body: str,
    attachments: list[dict] | None = None,
    cc: list[str] | None = None,
    bcc: list[str] | None = None,
) -> Email:
    """Sends an email with the given `body` to the given `address`. Returns a dictionary with the email details.

    :param recipients: The list with the email addresses of the recipients.
    :param subject: The subject of the email.
    :param body: The body of the email.
    :param attachments: The list of attachments to include in the email. If `null`, no attachments are included.
    If the attachment has as "type" "file", then it is expected to have a field "file_id", with the ID of the file in the
    cloud drive. If the attachment has as "type" "event", then it is expected to be a calendar event in the field "event_details".
    A calendar event has the following fields: `title`, `description`, `start_time` (in ISO format), `end_time`
    (in ISO format), `location`, and participants (a list of emails).
    :param cc: The list of email addresses to include in the CC field. If `null`, no email addresses are included.
    :param bcc: The list of email addresses to include in the BCC field. If `null`, no email addresses are included.
    """
    if attachments is not None:
        parsed_attachments = []
        for attachment in attachments:
            if not isinstance(attachment, dict):
                raise ValueError("Attachments must be dictionaries.")
            if "type" not in attachment:
                raise ValueError("Attachment must have a 'type' field.")
            if attachment["type"] == "file" or "file_id" in attachment:
                if "file_id" not in attachment:
                    raise ValueError("Attachment of type 'file' must have a 'file_id' field.")
                parsed_attachments.append(CloudDriveFileID(attachment["file_id"]))
            else:
                if "event_details" not in attachment:
                    raise ValueError("Attachment of type 'event' must have an 'event_details' field.")
                parsed_attachments.append(CalendarEvent(**attachment["event_details"]))
    else:
        parsed_attachments = None
    return inbox.send_email(recipients, subject, body, parsed_attachments, cc, bcc)


def search_emails(
    inbox: Annotated[Inbox, Depends("inbox")],
    query: str,
    sender: str | None = None,
) -> list[Email]:
    """Searches for emails in the inbox that contain the given query in the subject or body. If `address` is provided,
    only emails from that address are searched.

    :param query: The query to search for in the email subject or body. If empty, all emails are returned.
    :param sender: The email address of the sender. If `null`, all emails are searched. Use `search_contacts_by_name`
    if the email address is unknown.
    """
    matches = inbox.search_emails(query, sender)
    if len(matches) == 0:
        raise ValueError("No emails found. Try with a different query.")
    return matches


def delete_email(inbox: Annotated[Inbox, Depends("inbox")], email_id: str) -> str:
    """Deletes the email with the given `email_id` from the inbox.

    :param email_id: The id of the email to delete.
    """
    inbox.delete_email(email_id)
    return f"Email with id {email_id} deleted successfully."


def get_sent_emails(inbox: Annotated[Inbox, Depends("inbox")]) -> list[Email]:
    """Returns all the sent emails in the inbox. Each email has a recipient, a subject, and a body."""
    return inbox.sent


def get_received_emails(inbox: Annotated[Inbox, Depends("inbox")]) -> list[Email]:
    """Returns all the received emails in the inbox. Each email has a sender, a subject, and a body."""
    return inbox.received


def get_draft_emails(inbox: Annotated[Inbox, Depends("inbox")]) -> list[Email]:
    """Returns all the draft emails in the inbox. Each email has a recipient, a subject, and a body."""
    return inbox.drafts


def search_contacts_by_name(inbox: Annotated[Inbox, Depends("inbox")], query: str) -> list[EmailContact]:
    """Finds contacts in the inbox's contact list by name.
    It returns a list of contacts that match the given name.

    :param query: The name of the contacts to search for.
    """
    return inbox.find_contacts_by_name(query)


def search_contacts_by_email(inbox: Annotated[Inbox, Depends("inbox")], query: str) -> list[EmailContact]:
    """Finds contacts in the inbox's contact list by email.
    It returns a list of contacts that match the given email.

    :param query: The email of the contacts to search for.
    """
    return inbox.find_contacts_by_email(query)
